#include <iostream>
#include <blt/std/allocator.h>
#include <blt/gfx/window.h>
#include <blt/gfx/state.h>
#include <blt/gfx/stb/stb_image.h>
#include <blt/gfx/stb/stb_perlin.h>
#include <blt/std/hashmap.h>
#include <image.h>
#include "blt/gfx/renderer/resource_manager.h"
#include "blt/gfx/renderer/batch_2d_renderer.h"
#include "blt/std/assert.h"
#include "imgui.h"
#include <variant>
#include <random>

blt::gfx::matrix_state_manager global_matrices;
blt::gfx::resource_manager resources;
blt::gfx::batch_renderer_2d renderer_2d(resources);

struct node;

blt::area_allocator<node, 32000> node_allocator;
blt::area_allocator<image, 32> img_allocator;

using data_t = std::array<float, 5>;

node* createNode()
{
    auto* n = node_allocator.allocate(1);
    node_allocator.construct(n);
    return n;
}

image* createImage()
{
    auto* i = img_allocator.allocate(1);
    img_allocator.construct(i);
    return i;
}

void destroyNode(node* n)
{
    if (!n)
        return;
    node_allocator.destroy(n);
    node_allocator.deallocate(n, 1);
}

void destroyImage(image* img)
{
    if (!img)
        return;
    img_allocator.destroy(img);
    img_allocator.deallocate(img, 1);
}

class tree;

struct node
{
        friend tree;
    private:
        node* left = nullptr;
        node* right = nullptr;
        function_t type = function_t::ADD;
        data_t data{};
        image* img = nullptr;
        
        [[nodiscard]] bool has_both() const
        {
            return left->img != nullptr && right->img != nullptr;
        }
        
        [[nodiscard]] bool has_one() const
        {
            return left->img != nullptr || right->img != nullptr;
        }
        
        [[nodiscard]] image* get_one() const
        {
            if (left->img != nullptr)
                return left->img;
            if (right->img != nullptr)
                return right->img;
            BLT_THROW(std::runtime_error("Unable to get when one is missing! You have an error in your GA!"));
        }
    
    public:
        
        static node* construct_random_tree(const size_t depth = 0)
        {
            node* n = createNode();
            
            std::random_device dev;
            std::mt19937_64 engine{dev()};
            std::uniform_real_distribution<float> dist(0.0f, 1.0f);
            for (float& f : n->data)
                f = dist(engine);
            
            std::uniform_int_distribution op(0, OPERATOR_COUNT - 1);
            do
            {
                n->type = static_cast<function_t>(op(engine));
            } while (depth > 25 && inputs(n->type) > 0);
            
            switch (inputs(n->type))
            {
                case 2:
                    n->left = construct_random_tree(depth + 1);
                    n->right = construct_random_tree(depth + 1);
                    break;
                case 1:
                    n->left = construct_random_tree(depth + 1);
                    break;
                case 0:
                    break;
            }
            
            return n;
        }
        
        void evaluate()
        {
            destroyImage(img);
            img = createImage();
            for (i32 y = 0; y < height; y++)
            {
                for (i32 x = 0; x < width; x++)
                {
                    switch (type)
                    {
                        case function_t::ADD:
                        {
                            BLT_ASSERT(has_both());
                            img->set((left->img->get(x, y) + right->img->get(x, y)).normalize(), x, y);
                            break;
                        }
                        case function_t::SUB:
                        {
                            BLT_ASSERT(has_both());
                            img->set((left->img->get(x, y) - right->img->get(x, y)).normalize(), x, y);
                            break;
                        }
                        case function_t::MUL:
                        {
                            BLT_ASSERT(has_both());
                            img->set((left->img->get(x, y) * right->img->get(x, y)).normalize(), x, y);
                            break;
                        }
                        case function_t::DIV:
                        {
                            BLT_ASSERT(has_both());
                            //const auto& nem = left->img->get(x, y);
                            //const auto& den = right->img->get(x, y);
//                            auto v_x = den.x() == 0 ? 0 : nem.x() / den.x();
//                            auto v_y = den.y() == 0 ? 0 : nem.y() / den.y();
//                            auto v_z = den.z() == 0 ? 0 : nem.z() / den.z();
//                            img->set(blt::vec3{v_x, v_y, v_z}, x, y);
                            img->set(blt::vec3{0, 0, 0}.normalize(), x, y);
                            break;
                        }
                        case function_t::EXP:
                        {
                            BLT_ASSERT(has_both());
                            const auto& b = left->img->get(x, y);
                            const auto& e = right->img->get(x, y);
                            img->set(blt::vec3{std::pow(b.x(), e.x()), std::pow(b.y(), e.y()), std::pow(b.z(), e.z())}.normalize(), x, y);
                            break;
                        }
                        case function_t::LOG:
                        {
                            BLT_ASSERT(has_one());
                            const auto& v = get_one()->get(x, y);
                            auto v_x = v.x() == 0 ? 0.000001f : v.x();
                            auto v_y = v.y() == 0 ? 0.000001f : v.y();
                            auto v_z = v.z() == 0 ? 0.000001f : v.z();
                            img->set(blt::vec3{std::log(v_x), std::log(v_y), std::log(v_z)}.normalize(), x, y);
                            break;
                        }
                        case function_t::SQRT:
                        {
                            BLT_ASSERT(has_one());
                            const auto& v = get_one()->get(x, y);
                            img->set(blt::vec3{std::sqrt(v.x()), std::sqrt(v.y()), std::sqrt(v.z())}.normalize(), x, y);
                            break;
                        }
                        case function_t::QUAD:
                        {
                            BLT_ASSERT(has_one());
                            auto v = get_one()->get(x, y);
                            img->set(v * v, x, y);
                            break;
                        }
                        case function_t::RANDOM:
                        {
                            static std::random_device dev;
                            static std::mt19937_64 engine(dev());
                            static std::uniform_real_distribution<float> dist(0, 1);
                            img->set(blt::vec3{dist(engine), dist(engine), dist(engine)}, x, y);
                            break;
                        }
                        case function_t::NOISE:
                        {
                            auto v = std::numeric_limits<u8>::max() *
                                     stb_perlin_noise3(static_cast<float>(x) / ((256.0f * (data[0] + 1.0f) / 2.0f)),
                                                       static_cast<float>(y) / ((256 * (data[1] + 1) / 2)), data[2] / 2.0f + 0.5f, 0,
                                                       0, 0);
                            img->set(blt::vec3{v, v, v}.normalize(), x, y);
                            break;
                        }
                        case function_t::COLOR:
                        {
                            img->set(blt::vec3{data[0], data[1], data[2]}.normalize(), x, y);
                            break;
                        }
                        case function_t::SCALAR:
                        {
                            img->set(blt::vec3{data[0], data[0], data[0]}.normalize(), x, y);
                            break;
                        }
                    }
                }
            }
        }
        
        void evaluate_tree()
        {
            if (left)
                left->evaluate_tree();
            if (right)
                right->evaluate_tree();
            evaluate();
        }
        
        image* getImage()
        {
            return img;
        }
        
        ~node()
        {
            destroyImage(img);
            destroyNode(left);
            destroyNode(right);
        }
};

class tree
{
    public:
        node* root;
    
    
};

node* root;
blt::gfx::texture_gl2D* texture;

bool fucky(float f)
{
    return f == std::numeric_limits<float>::quiet_NaN() || f == -std::numeric_limits<float>::quiet_NaN() || f == std::numeric_limits<float>::infinity() ||
           f == -std::numeric_limits<float>::infinity();
}

void init()
{
    global_matrices.create_internals();
    renderer_2d.create();
    
    
    texture = new blt::gfx::texture_gl2D(width, height);
    resources.set("img", texture);
    resources.enqueue("../libraries/BLT-With-Graphics-Template/resources/textures/parkerfemBOY.png", "cum");
    
    resources.load_resources();
}

void update(std::int32_t w, std::int32_t h)
{
    global_matrices.update_perspectives(w, h, 90, 0.1, 2000);
    glDisable(GL_BLEND);
    
    if (ImGui::Button("Regenerate"))
    {
        BLT_INFO("Regen tree");
        destroyNode(root);
        root = node::construct_random_tree();
        root->evaluate_tree();
        BLT_INFO("Preprocess");
        auto& data = root->getImage()->getData();
        for (auto& v : data)
        {
            //BLT_INFO_STREAM << "B4: " << v.x() << " " << v.y() << " " << v.z() << " " << v.w() << "\n";
            if (fucky(v.x()))
                v[0] = 0;
            if (fucky(v.y()))
                v[1] = 0;
            if (fucky(v.z()))
                v[2] = 0;
            v = v.normalize();
            if (fucky(v.x()))
                v[0] = 0;
            if (fucky(v.y()))
                v[1] = 0;
            if (fucky(v.z()))
                v[2] = 0;
            //BLT_INFO_STREAM << "A4: " << v.x() << " " << v.y() << " " << v.z() << " " << v.w() << "\n";
        }
        
        BLT_INFO("Uploading");
        texture->upload((void*) root->getImage()->getData().data(), GL_RGB, 0, 0, 0, -1, -1, GL_FLOAT);
        resources.set("img", texture);
    }
    
    auto lw = 512.0f;
    auto lh = 512.0f;
    //renderer_2d.drawRectangle(blt::vec4{0.5, 0.0, 1.0, 1.0},
    //                          {static_cast<float>(w) / 2.0f, static_cast<float>(h) / 2.0f, static_cast<float>(w), static_cast<float>(h)});
    renderer_2d.drawRectangle("cum", {static_cast<float>(w) / 2.0f, static_cast<float>(h) / 2.0f, lw, lh});
    
    global_matrices.update();
    renderer_2d.render();
}

int main()
{
    blt::gfx::init(blt::gfx::window_data{"Window of GP test", init, update}.setSyncInterval(1));
    global_matrices.cleanup();
    resources.cleanup();
    renderer_2d.cleanup();
    blt::gfx::cleanup();
    return 0;
}
